#!/usr/bin/env python3

import radical.entk  as re

if __name__ == '__main__':

    pipelines = set()
    for i in range(128):
        t = re.Task()
        t.executable = '/bin/date'

        s = re.Stage()
        s.add_tasks(t)

        p = re.Pipeline()
        p.add_stages(s)

        pipelines.add(p)


    amgr = re.AppManager(autoterminate=True, hostname='localhost', port=5672)
    amgr.resource_desc = {'resource': 'local.localhost',
                          'walltime': 10,
                          'cpus'    : 4}

    amgr.workflow = set(pipelines)
    amgr.run()
    amgr.terminate()





__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__    = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__license__   = "MIT"


import os
import json
import pika
import queue

import threading       as mt
import multiprocessing as mp

import radical.utils as ru

from ...exceptions       import EnTKError
from ...                 import states, Task
from ..base.task_manager import Base_TaskManager


# pylint: disable=unused-argument
# ------------------------------------------------------------------------------
#
class TaskManager(Base_TaskManager):
    """
    A Task Manager takes the responsibility of dispatching tasks it receives
    from a pending_queue for execution on to the available resources using a
    runtime system. Once the tasks have completed execution, they are pushed
    on to the completed_queue for other components of EnTK to process.

    :arguments:
        :pending_queue:     (list) List of queue(s) with tasks ready to be
                            executed. Currently, only one queue.
        :completed_queue:   (list) List of queue(s) with tasks that have
                            finished execution. Currently, only one queue.
        :rmgr:              (ResourceManager) Object to be used to access the
                            Pilot where the tasks can be submitted
        :rmq_conn_params:   (pika.connection.ConnectionParameters) object of
                            parameters necessary to connect to RabbitMQ

    Currently, EnTK is configured to work with one pending queue and one
    completed queue. In the future, the number of queues can be varied for
    different throughput requirements at the cost of additional Memory and CPU
    consumption.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, sid, pending_queue, completed_queue, rmgr,
                       rmq_conn_params):

        super(TaskManager, self).__init__(sid, pending_queue, completed_queue,
                                          rmgr, rmq_conn_params, rts='mock')
        self._rts_runner = None

        self._rmq_ping_interval = int(os.getenv('RMQ_PING_INTERVAL', '10'))

        self._log.info('Created task manager object: %s', self._uid)
        self._prof.prof('tmgr_create', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _tmgr(self, uid, rmgr, pending_queue, completed_queue,
                    rmq_conn_params):
        """
        **Purpose**: Method to be run by the tmgr process. This method receives
                     a Task from the pending_queue and submits it to the RTS.
                     Currently, it also converts Tasks into CUDs and CUs into
                     (partially described) Tasks.  This conversion is necessary
                     since the current RTS is RADICAL Pilot.  Once Tasks are
                     recovered from a CU, they are then pushed to the
                     completed_queue. At all state transititons, they are synced
                     (blocking) with the AppManager in the master process.

                     In addition the tmgr also receives heartbeat 'request' msgs
                     from the heartbeat-req queue. It responds with a 'response'
                     message to the 'heartbeart-res' queue.

        **Details**: The AppManager can re-invoke the tmgr process with this
                    function if the execution of the workflow is still
                    incomplete. There is also population of a dictionary,
                    `placeholders`, which stores the path of each of the tasks
                    on the remote machine.
        """

        try:

            # ------------------------------------------------------------------
            def heartbeat_response(mq_channel):

                try:

                    # Get request from heartbeat-req for heartbeat response
                    method_frame, props, body = \
                                  mq_channel.basic_get(queue=self._hb_request_q)

                    if not body:
                        return

                    self._log.info('Received heartbeat request')

                    nprops = pika.BasicProperties(
                                            correlation_id=props.correlation_id)
                    mq_channel.basic_publish(
                                exchange='',
                                routing_key=self._hb_response_q,
                                properties=nprops,
                                body='response')

                    self._log.info('Sent heartbeat response')

                    mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)


                except Exception as e:
                    self._log.exception('Failed to respond to heartbeat, '
                                        'error: %s', e)
                    raise
            # ------------------------------------------------------------------

            self._prof.prof('tmgr process started', uid=self._uid)
            self._log.info('Task Manager process started')

            # Acquire a connection+channel to the rmq server
            mq_connection = pika.BlockingConnection(rmq_conn_params)
            mq_channel = mq_connection.channel()

            # Make sure the heartbeat response queue is empty
            mq_channel.queue_delete(queue=self._hb_response_q)
            mq_channel.queue_declare(queue=self._hb_response_q)

            # Queue for communication between threads of this process
            task_queue = queue.Queue()

            # Start second thread to receive tasks and push to RTS
            self._rts_runner = mt.Thread(target=self._process_tasks,
                                         args=(task_queue, rmgr,
                                               rmq_conn_params))
            self._rts_runner.start()

            self._prof.prof('tmgr infrastructure setup done', uid=uid)

            while not self._tmgr_terminate.is_set():

                try:

                    # Get tasks from the pending queue
                    method_frame, _, body = \
                                    mq_channel.basic_get(queue=pending_queue[0])

                    if body:

                        body = json.loads(body)
                        task_queue.put(body)

                        mq_channel.basic_ack(
                                delivery_tag=method_frame.delivery_tag)

                    heartbeat_response(mq_channel)

                    # Raise an exception while running tests
                    ru.raise_on(tag='tmgr_fail')

                except Exception as e:
                    self._log.exception('Error in task execution: %s', e)
                    raise


        except KeyboardInterrupt:

            self._log.exception('Execution interrupted (probably by Ctrl+C), '
                                'cancel tmgr process gracefully...')
            raise


        except Exception as e:

            self._log.exception('%s failed with %s', self._uid, e)
            raise EnTKError(e) from e

        finally:

            self._prof.prof('tmgr_term', uid=uid)

            if self._rts_runner:
                self._rts_runner.join()

            mq_connection.close()
            self._prof.close()


    # --------------------------------------------------------------------------
    #
    def _process_tasks(self, task_queue, rmgr, rmq_conn_params):
        '''
        **Purpose**: The new thread that gets spawned by the main tmgr process
                     invokes this function. This function receives tasks from
                     'task_queue' and submits them to the RADICAL Pilot RTS.
        '''

        # placeholders = dict()

        # # --------------------------------------------------------------------
        # def load_placeholder(task):
        # 
        #     parent_pipeline = str(task.parent_pipeline['name'])
        #     parent_stage = str(task.parent_stage['name'])
        # 
        #     if parent_pipeline not in placeholders:
        #         placeholders[parent_pipeline] = dict()
        # 
        #     if parent_stage not in placeholders[parent_pipeline]:
        #         placeholders[parent_pipeline][parent_stage] = dict()
        # 
        #     if None not in [parent_pipeline, parent_stage, task.name]:
        #         placeholders[parent_pipeline][parent_stage][str(
        #             task.name)] = str(task.path)
        # # --------------------------------------------------------------------

        mq_connection = pika.BlockingConnection(rmq_conn_params)
        mq_channel = mq_connection.channel()

        try:

            while not self._tmgr_terminate.is_set():

                body = None

                try:
                    body = task_queue.get(block=True, timeout=10)

                except queue.Empty:
                    # Ignore, we don't always have new tasks to run
                    pass

                if not body:
                    continue

                task_queue.task_done()

                bulk_tasks = list()

                for msg in body:

                    task = Task()
                    task.from_dict(msg)
                    bulk_tasks.append(task)

                    self._advance(task, 'Task', states.SUBMITTING,
                                  mq_channel, rmq_conn_params,
                                  '%s-tmgr-to-sync' % self._sid)

                # this mock RTS immmedialtely completes all tasks
                for task in bulk_tasks:

                    self._advance(task, 'Task', states.COMPLETED,
                                  mq_channel, rmq_conn_params,
                                  '%s-cb-to-sync' % self._sid)

                    task_as_dict = json.dumps(task.to_dict())
                    mq_channel.basic_publish(
                            exchange='',
                            routing_key='%s-completedq-1' % self._sid,
                            body=task_as_dict)

                    self._log.info('Pushed task %s with state %s to '
                                   'completed queue %s-completedq-1',
                                   task.uid, task.state, self._sid)

        except KeyboardInterrupt:
            self._log.exception('Execution interrupted (probably by Ctrl+C), '
                                'cancel task processor gracefully...')

        except Exception as e:
            self._log.exception('%s failed with %s', self._uid, e)
            raise EnTKError(e) from e


    # --------------------------------------------------------------------------
    #
    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr process. The tmgr function
                     is not to be accessed directly. The function is started
                     in a separate thread using this method.
        """

        if self._tmgr_process:
            self._log.warn('tmgr process already running!')
            return


        try:

            self._prof.prof('creating tmgr process', uid=self._uid)
            self._tmgr_terminate = mp.Event()

            self._tmgr_process = mp.Process(target=self._tmgr,
                                            name='task-manager',
                                            args=(self._uid,
                                                  self._rmgr,
                                                  self._pending_queue,
                                                  self._completed_queue,
                                                  self._rmq_conn_params)
                                            )

            self._log.info('Starting task manager process')
            self._prof.prof('starting tmgr process', uid=self._uid)

            self._tmgr_process.start()

            return True

        except Exception as e:

            self._log.exception('Task manager not started, error: %s', e)
            self.terminate_manager()
            raise


# ------------------------------------------------------------------------------




import os, json, time
from radical.entk import Pipeline, Stage, Task, AppManager

# ------------------------------------------------------------------------------
# Set default verbosity

if os.environ.get('RADICAL_ENTK_VERBOSE') is None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'

# Assumptions:
# - # of MD steps: 2
# - Each MD step runtime: 15 minutes
# - Summit's scheduling policy [1]
#
# Resource rquest:
# - 4 <= nodes with 2h walltime.
#
# Workflow [2]
#
# [1] https://www.olcf.ornl.gov/for-users/system-user-guides/summit/summit-user-guide/scheduling-policy
# [2] https://docs.google.com/document/d/1XFgg4rlh7Y2nckH0fkiZTxfauadZn_zSn3sh51kNyKE/
#
RMQ_HOSTNAME = 'two.radical-project.org'
RMQ_PORT     = 33235
'''
export RADICAL_PILOT_DBURL=mongodb://hyperrct:h1p3rrc7@two.radical-project.org:27017/hyperrct
export RADICAL_PILOT_PROFILE=True
export RADICAL_ENTK_PROFILE=True
'''
#


N_jobs_MD = 12
N_jobs_ML = 10

hrs_wt = 2  # walltime in hours
queue = 'batch'

CUR_STAGE = 0
MAX_STAGE = 0  # 10
RETRAIN_FREQ = 5

LEN_initial = 10  # 100 initial stage sim time in ns
LEN_iter = 10


def generate_training_pipeline():
    """
    Function to generate the CVAE_MD pipeline
    """

    def generate_MD_stage(num_MD=1):
        """
        Function to generate MD stage.
        """
        s1 = Stage()
        s1.name = 'MD'

        # MD tasks
        for i in range(num_MD):
            t1 = Task()
            t1.executable = ['/bin/sleep']
            t1.arguments  = ['3']

            # assign hardware the task
            t1.cpu_reqs = {'processes': 1,
                           'process_type': None,
                           'threads_per_process': 4,
                           'thread_type': 'OpenMP'
                          }
            t1.gpu_reqs = {'processes': 1,
                           'process_type': None,
                           'threads_per_process': 1,
                           'thread_type': 'CUDA'
                          }

            # Add the MD task to the simulating stage
            s1.add_tasks(t1)
        return s1


    def generate_aggregating_stage():
        """
        Function to concatenate the MD trajectory (h5 contact map)
        """
        s2 = Stage()
        s2.name = 'aggregating'

        # Aggregation task
        t2 = Task()
        t2.executable = ['/bin/sleep']
        t2.arguments = ['3']

        # Add the aggregation task to the aggreagating stage
        s2.add_tasks(t2)
        return s2


    def generate_ML_stage(num_ML=1):
        """
        Function to generate the learning stage
        """
        s3 = Stage()
        s3.name = 'learning'

        # learn task
        time_stamp = int(time.time())
        for i in range(num_ML):
            t3 = Task()
            t3.executable = ['/bin/sleep']
            t3.arguments = ['3']
            t3.cpu_reqs = {'processes': 1,
                           'process_type': None,
                           'threads_per_process': 4,
                           'thread_type': 'OpenMP'
                          }
            t3.gpu_reqs = {'processes': 1,
                           'process_type': None,
                           'threads_per_process': 1,
                           'thread_type': 'CUDA'
                          }

            # Add the learn task to the learning stage
            s3.add_tasks(t3)

        # TICA jobs
        for i in range(num_ML):
            t3 = Task()
            t3.executable = ['/bin/sleep']
            t3.arguments = ['3']

            t3.cpu_reqs = {'processes': 1,
                           'process_type': None,
                           'threads_per_process': 4,
                           'thread_type': 'OpenMP'
                          }
            # Add the learn task to the learning stage
            s3.add_tasks(t3)

        return s3


    def generate_interfacing_stage():
        s4 = Stage()
        s4.name = 'scanning'

        # Scaning for outliers and prepare the next stage of MDs
        t4 = Task()
        t4.executable = ['/bin/sleep']
        t4.arguments = ['3']

        t4.cpu_reqs = {'processes': 1,
                       'process_type': None,
                       'threads_per_process': 12,
                       'thread_type': 'OpenMP'
                      }
        t4.gpu_reqs = {'processes': 1,
                       'process_type': None,
                       'threads_per_process': 1,
                       'thread_type': 'CUDA'
                      }
        s4.add_tasks(t4)
        s4.post_exec = func_condition

        return s4


    def func_condition():

        global CUR_STAGE, MAX_STAGE
        if CUR_STAGE < MAX_STAGE:
            func_on_true()
        else:
            func_on_false()


    def func_on_true():
        global CUR_STAGE, MAX_STAGE
        print ('finishing stage %d of %d' % (CUR_STAGE, MAX_STAGE) )

        # --------------------------
        # MD stage
        s1 = generate_MD_stage(num_MD=N_jobs_MD)
        # Add simulating stage to the training pipeline
        p.add_stages(s1)

        if CUR_STAGE % RETRAIN_FREQ == 0:
            # --------------------------
            # Aggregate stage
            s2 = generate_aggregating_stage()
            # Add the aggregating stage to the training pipeline
            p.add_stages(s2)

            # --------------------------
            # Learning stage
            s3 = generate_ML_stage(num_ML=N_jobs_ML)
            # Add the learning stage to the pipeline
            p.add_stages(s3)

        # --------------------------
        # Outlier identification stage
        s4 = generate_interfacing_stage()
        p.add_stages(s4)

        CUR_STAGE += 1

    def func_on_false():
        print ('Done' )


    global CUR_STAGE

    p = Pipeline()
    p.name = 'MD_ML'

    func_on_true()

    return p


if __name__ == '__main__':

    # Create a dictionary to describe four mandatory keys:
    # resource, walltime, cores and project
    # resource is 'local.localhost' to execute locally
    res_dict = {
            'resource': 'local.localhost',
          # 'queue'   : queue,
            'schema'  : 'local',
            'walltime': 60 * hrs_wt,
            'cpus'    : N_jobs_MD * 7,
            'gpus'    : N_jobs_MD,  # 6 * 2,

          # 'resource': 'ornl.summit',
          # 'queue'   : queue,
          # 'schema'  : 'local',
          # 'walltime': 60 * hrs_wt,
          # 'cpus'    : N_jobs_MD * 7,
          # 'gpus'    : N_jobs_MD,  # 6 * 2,
          # 'project' : 'LRN005',
    }

    # Create Application Manager
    # appman = AppManager()
    appman = AppManager(hostname=RMQ_HOSTNAME, port=RMQ_PORT)
    appman.resource_desc = res_dict

    p1 = generate_training_pipeline()

    appman.workflow = [p1]

    appman.run()

